package com.loafle.overflow.central.proxy;

import com.loafle.overflow.central.api.CentralAPIGrpc;
import com.loafle.overflow.central.api.ServerInput;
import com.loafle.overflow.central.api.ServerOutput;
import com.loafle.overflow.central.spring.AppConfig;
import com.loafle.overflow.core.exception.OverflowException;

import io.grpc.*;
import io.grpc.Status.Code;
import io.grpc.netty.NettyServerBuilder;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.util.logging.Logger;

/**
 * Created by insanity on 17. 6. 28.
 */
public class ServiceProxy {
  private static final Logger logger = Logger.getLogger(ServiceProxy.class.getName());

  private io.grpc.Server server;
  private ApplicationContext ctx;

  public void start(int port) throws IOException {
    ctx = new AnnotationConfigApplicationContext(AppConfig.class);

    ProxyServerInterceptor proxyServerInterceptor = new ProxyServerInterceptor();
    ServiceImpl serviceImpl = ctx.getBean(ServiceImpl.class);

    server = NettyServerBuilder.forPort(port)
        .addService(ServerInterceptors.intercept(serviceImpl, proxyServerInterceptor))
        .build().start();
    logger.info("Server started, listening on " + port);
    Runtime.getRuntime().addShutdownHook(new Thread() {
      @Override
      public void run() {
        // Use stderr here since the logger may have been reset by its JVM shutdown hook.
        System.err.println("*** shutting down gRPC server since JVM is shutting down");
        ServiceProxy.this.stop();
        System.err.println("*** server shut down");
      }
    });
  }

  public void stop() {
    if (server != null) {
      server.shutdown();
    }
  }

  public void blockUntilShutdown() throws InterruptedException {
    if (server != null) {
      server.awaitTermination();
    }
  }

  @Component
  static class ServiceImpl extends CentralAPIGrpc.CentralAPIImplBase {

    @Autowired
    private ServiceInvoker serviceInvoker;

    ServiceImpl(ServiceInvoker serviceInvoker) {
      this.serviceInvoker = serviceInvoker;
    }

    @Override
    public void exec(ServerInput request, io.grpc.stub.StreamObserver<ServerOutput> responseObserver) {

      String result = null;

      try {
        result = this.serviceInvoker.invoke(request.getTarget(), request.getMethod(),
            request.getParamsList().asByteStringList());

        ServerOutput reply = ServerOutput.newBuilder().setResult(result).build();
        responseObserver.onNext(reply);
        responseObserver.onCompleted();

      } catch (OverflowException e) {
        responseObserver.onError(this.getException(request, Status.Code.INTERNAL, e));
        return;
      } catch (NoSuchMethodException e) {
        responseObserver.onError(this.getException(request, Status.Code.NOT_FOUND, e));
        return;
      } catch (IllegalAccessException e) {
        responseObserver.onError(this.getException(request, Status.Code.INVALID_ARGUMENT, e));
        return;
      }
    }

    private StatusException getException(ServerInput request, Code code, Throwable cause) {

      String msg = String.format("Target: %s, Method:%s, Params:%s, message:%s", request.getTarget(),
          request.getMethod(), request.getParamsList().asByteStringList(), cause.getMessage());
      Status status = Status.fromCode(code).withDescription(msg).withCause(cause);
      return new StatusException(status);
    }
  }

}

